#include <Arduino.h>

#include <TimerOne.h>
#include <Wire.h>

#include <Adafruit_MotorShield.h>
//#include "utility/Adafruit_MS_PWMServoDriver.h"

#define NUM_MOTORS 4  // ( 4 -> wheel motors)

// Create the motor shield object with the default I2C address
Adafruit_MotorShield AFMS = Adafruit_MotorShield(); 
// Or, create it with a different I2C address (say for stacking)
// Adafruit_MotorShield AFMS = Adafruit_MotorShield(0x61); 

// Select which 'port' M0, M1, M2 or M3. In this case, M1
Adafruit_DCMotor* M[NUM_MOTORS];
// You can also make another motor on port M2
//Adafruit_DCMotor *myOtherMotor = AFMS.getMotor(2);

#include "channels.h"
#include "proj_types.h"

// 9 -> Servo Pulse Left * 
// 10 -> Servo Pulse Right *

// PORTA
// 22 <- ENC0_A
// 23 <- ENC0_B
// 24 <- ENC1_A
// 25 <- ENC1_B
// 26 <- ENC2_A
// 27 <- ENC2_B
// 28 <- ENC3_A
// 29 <- ENC3_B

// 6 <- Button Start * 
// 7 <- Button Stop *

// A11 <- Lidar motor voltage

// Metal Gearmotor 12VDC with 64 CPR Encoder 350rpm
// Color  - Function

// Black  - motor power
// Red    - motor power
// Blue   - Hall sensor Vcc (3.5 ? 20 V)
// Green  - Hall sensor GND
// Yellow - Hall sensor A output
// White  - Hall sensor B output


#define TICKS_PER_REV 2880.0
#define TICKS_PER_RAD (TICKS_PER_REV / TWO_PI)
#define RADS_PER_TICK (TWO_PI / TICKS_PER_REV)

const int encoder0A_pin = 22;  // Motor0 encoder A input
const int encoder0B_pin = 23;  // Motor0 encoder B input
const int encoder1A_pin = 24;  // Motor1 encoder A input
const int encoder1B_pin = 25;  // Motor1 encoder B input
const int encoder2A_pin = 26;  // Motor2 encoder A input
const int encoder2B_pin = 27;  // Motor2 encoder B input
const int encoder3A_pin = 28;  // Motor3 encoder A input
const int encoder3B_pin = 29;  // Motor3 encoder B input

channels_t serial_channels;


uint8_t encoder_state[NUM_MOTORS];
volatile int16_t encoder_delta[NUM_MOTORS];
uint32_t encoder_pos[NUM_MOTORS];
uint16_t motor_timeout_count;

int M_PWM_req[NUM_MOTORS];
int M_PWM_act[NUM_MOTORS];
int delta_pw_max;

PID_t pid[NUM_MOTORS];
int odo[NUM_MOTORS];

float dt;
float w_odo[NUM_MOTORS];
float w_ref[NUM_MOTORS];
int PWM_out[NUM_MOTORS];

int button_stop, prev_botton_stop;

//#pragma GCC push_options
//#pragma GCC optimize ("O3")

void timer_interrupt(void)
{
  uint8_t b, new_state;
  static int8_t encoder_table[16] = {0, 1, -1, 0,  -1, 0, 0, 1,  1, 0, 0, -1,  0, -1, 1, 0};
  
  b = PINA;
  
  // Put encoder 0 channels in the lowest bits
  //new_state = (digitalRead(encoder0A_pin) << 1) | digitalRead(encoder0B_pin);
  new_state = b & 0b00000011;
  encoder_delta[0] += encoder_table[encoder_state[0] | new_state];
  encoder_state[0] = new_state << 2;

  // Put encoder 1 channels in the lowest bits
  //new_state = (digitalRead(encoder1A_pin) << 1) | digitalRead(encoder1B_pin);
  b = (b >> 2);
  new_state = b & 0b00000011;
  encoder_delta[1] += encoder_table[encoder_state[1] | new_state];
  encoder_state[1] = new_state << 2;

  // Put encoder 2 channels in the lowest bits
  //new_state = (digitalRead(encoder2A_pin) << 1) | digitalRead(encoder2B_pin);
  b = (b >> 2);
  new_state = b & 0b00000011;
  encoder_delta[2] += encoder_table[encoder_state[2] | new_state];
  encoder_state[2] = new_state << 2;
  
  // Put encoder 3 channels in the lowest bits
  //new_state = (digitalRead(encoder3A_pin) << 1) | digitalRead(encoder3B_pin);
  b = (b >> 2);
  new_state = b & 0b00000011;
  encoder_delta[3] += encoder_table[encoder_state[3] | new_state];
  encoder_state[3] = new_state << 2;
}

//#pragma GCC pop_options

const int led_pin =  13;     // the number of the LED pin

const int button_start_pin = 6;
const int button_stop_pin = 7;

const int servo_left_pin = 9;
const int servo_right_pin = 10;


int servo_state = 0;

int servo_left_pulse = 1000;
int servo_right_pulse = 1500;
/*
void servo_interrupt(void)
{
  if (servo_state == 0) {
    Timer4.start(servo_left_pulse);
    digitalWrite(servo_left_pin, 1);  // Start Left Servo Pulse
    servo_state = 1;
  } else if (servo_state == 1) {
    digitalWrite(servo_left_pin, 0); // End Left Servo Pulse
    Timer4.start(servo_right_pulse);
    digitalWrite(servo_right_pin, 1);  // Start Right Servo Pulse
    servo_state = 2;
  } else if (servo_state == 2) {
    digitalWrite(servo_right_pin, 0); // End right Servo Pulse
    Timer4.start(20000 - (servo_left_pulse + servo_right_pulse));
    servo_state = 0;
  }
}
*/

int led_state = 0;  
uint32_t last_micros = 0;  
uint32_t last_motor_update = 0;  

uint32_t interval;         // interval at which to blink (microseconds)


void swap(int& a, int& b)
{
 int t = a;
 a = b;
 b = t;
}


void setup_PIDs(void)
{
  pid[0].Kp = 0.03;
  pid[0].Ki = 0.3;
  pid[0].Kd = 0;
  pid[0].Kf = 0.0375;
  pid[0].scale = 255;
  pid[0].dt = 0.04;
  pid[0].active = 1;

  pid[1] = pid[0];
  pid[2] = pid[0];
  pid[3] = pid[0];
}

float calc_pid(PID_t& pid, float w_ref, float w)
{
  float de, Se;
  pid.last_e = pid.e;
  pid.e = w_ref - w;
  pid.w = w;
  pid.w_ref = w_ref;

  Se = pid.Se + pid.e * pid.dt;
  de = (pid.e - pid.last_e) / pid.dt;

  // Remove integration for zero reference
  if (w_ref == 0) {
    Se = 0;
  }

  // Calc PID output
  pid.m = pid.Kp * pid.e + pid.Ki * Se + pid.Kd * de + pid.Kf * pid.w_ref;

  // Anti windup
  if (abs(pid.m) > 1) {
    pid.m = pid.m + pid.Ki * (pid.Se - Se);
  } else {
    pid.Se = Se;
  }

  if (pid.m > 1) {
    pid.m = 1;
  } else if (pid.m < -1) {
    pid.m = -1;
  }

  return pid.m * pid.scale;
}

void setMotorPWM(uint8_t num, int new_PWM, int mot_enable)
{
  if (num >= NUM_MOTORS) return;
  
  if (new_PWM > 255) new_PWM = 255;
  if (new_PWM < -255) new_PWM = -255;
  
  M_PWM_act[num] = new_PWM;
  //return;

  if (mot_enable) {
    if (new_PWM >= 0) {
      M[num]->setSpeed(new_PWM);
      M[num]->run(FORWARD); 
    } else {
      M[num]->setSpeed(abs(new_PWM));
      M[num]->run(BACKWARD);
    }
  } else {
    M[num]->run(RELEASE);
  }
}

void setMotorReqPWM(uint8_t num, int new_PWM)
{
  if (num >= NUM_MOTORS) return;
  
  if (new_PWM > 255) new_PWM = 255;
  if (new_PWM < -255) new_PWM = -255;
  
  M_PWM_req[num] = new_PWM;
}

void updateMotorPWM(uint8_t num)
{
  int delta_pwm;
  if (num >= NUM_MOTORS) return;
  
  delta_pwm = M_PWM_req[num] - M_PWM_act[num];
  if (delta_pwm >  delta_pw_max) delta_pwm =  delta_pw_max;
  if (delta_pwm < -delta_pw_max) delta_pwm = -delta_pw_max;
  
  M_PWM_act[num] += delta_pwm;
}


// G
// H
// I - dt
// J
// K
// L
// M - Motors PWM
// N
// O
// P - PID Active
// Q - Set Encoder value
// R
// S - Servo Pulses
// T
// U
// V
// W
// X
// Y
// Z


// g - ping++
// h -
// i -
// j -
// k -
// l - MOTOR PID Kp
// m - MOTOR PID Ki
// n - MOTOR PID Kd
// o - MOTOR PID Kf
// p
// q
// r - W0 ref
// s - W1 ref
// t - W2 ref
// u - W3 ref
// v
// w
// x
// y
// z


void process_serial_packet(char channel, uint32_t value, channels_t& obj)
{
 byte i;

 if (channel == 'G') {

 } else if (channel == 'S')  {
   servo_left_pulse = (value >> 16) & 0xFFFF;
   servo_right_pulse = value & 0xFFFF;
   
 } else if (channel == 'M')  {
   int16_t pwm = value & 0xFFFF;
   uint8_t num_motor = (value >> 24) & 0x03;
   setMotorReqPWM(num_motor, pwm);
   motor_timeout_count = 0;
   for (i = 0; i < NUM_MOTORS; i++){
     pid[i].active = 0;
   }

 } else if (channel == 'P')  {
   for (i = 0; i < NUM_MOTORS; i++){
     pid[i].active = value;
   }

 } else if (channel == 'Q')  {
   for (i = 0; i < NUM_MOTORS; i++) {
     encoder_pos[i] = value;
   }

 } else if (channel == 'r' ||
            channel == 's' ||
            channel == 't' ||
            channel == 'u')  {

   uint8_t num_motor = channel - 'r';
   w_ref[num_motor] = *((float*)(&value));
   for (i = 0; i < NUM_MOTORS; i++){
     pid[i].active = 1;
   }
 
 } else if (channel == 'l')  {
   for (i = 0; i < NUM_MOTORS; i++)
     pid[i].Kp = *((float*)(&value));

 } else if (channel == 'm')  {
   for (i = 0; i < NUM_MOTORS; i++) {
     pid[i].Ki = *((float*)(&value));
     pid[i].Se = 0;
   }
 
 } else if (channel == 'n')  {
   for (i = 0; i < NUM_MOTORS; i++)
     pid[i].Kd = *((float*)(&value));

 } else if (channel == 'o')  {
   for (i = 0; i < NUM_MOTORS; i++)
     pid[i].Kf = *((float*)(&value));

 } else if (channel == 'I')  {
   interval = value;

 } else if (channel == 'g')  { // Ping
   obj.send(channel, value + 1);
   Serial.println(value + 1);
 }
}



void serial_write(uint8_t b)
{
  Serial.write(b);
}


void setup(void)
{
  uint8_t i;
  // set the digital pin as output:
  pinMode(led_pin, OUTPUT);      
  digitalWrite(led_pin, 0);      

  digitalWrite(servo_left_pin, 0);      
  pinMode(servo_left_pin, OUTPUT);      

  digitalWrite(servo_right_pin, 0);      
  pinMode(servo_right_pin, OUTPUT);      

  pinMode(button_start_pin, INPUT_PULLUP);
  pinMode(button_stop_pin, INPUT_PULLUP);
  button_stop = !digitalRead(button_stop_pin);
  prev_botton_stop = button_stop;

  pinMode(encoder0A_pin, INPUT_PULLUP);
  pinMode(encoder0B_pin, INPUT_PULLUP);
  pinMode(encoder1A_pin, INPUT_PULLUP);
  pinMode(encoder1B_pin, INPUT_PULLUP);
  pinMode(encoder2A_pin, INPUT_PULLUP);
  pinMode(encoder2B_pin, INPUT_PULLUP);
  pinMode(encoder3A_pin, INPUT_PULLUP);
  pinMode(encoder3B_pin, INPUT_PULLUP);
  

  interval = 40000UL;
  delta_pw_max = 64;
  
  //pinMode(PWM_pin, OUTPUT);
  //analogWrite(PWM_pin, PWM_out);   

  timer_interrupt();
  for(i = 0; i < NUM_MOTORS; i++) {
    encoder_delta[i] = 0;
  }

  Timer1.attachInterrupt(timer_interrupt);
  Timer1.initialize(50);

  
  //Timer3.attachInterrupt(timer_interrupt);
  //Timer3.start(50); // Calls every 50us

  //Timer4.attachInterrupt(servo_interrupt);
  //Timer4.start(20000); // Calls every 20 ms

  // initialize serial communication:
  Serial.begin(115200);
  serial_channels.init(process_serial_packet, serial_write);
  Serial.println("Serial Init");
 
    
  AFMS.begin();  // create with the default frequency 1.6KHz
  //AFMS.begin(1000);  // OR with a different frequency, say 1KHz

  for (i = 0; i < NUM_MOTORS; i++) {
    M[i] = AFMS.getMotor(i + 1);
  }
  
  //M[0]->setSpeed(150);
  //M[0]->run(BACKWARD); 
  //M[0]->run(FORWARD); 

  setMotorPWM(0, 0, !button_stop); 
  setMotorPWM(1, 0, !button_stop); 
  setMotorPWM(2, 0, !button_stop); 
  setMotorPWM(3, 0, !button_stop); 
    

  setup_PIDs();

  last_micros = micros();
  last_motor_update = micros();
  Serial.println("Init Done");
  

}


void loop()
{
  byte i; 
  static int decimate_led  = 0;
  int32_t act_micros, delta;

  if (Serial.available() > 0) {
    // read the oldest byte in the serial buffer:
    byte serialByte = Serial.read();
    serial_channels.StateMachine(serialByte);
    //Serial.write(serialByte);
  }

  //digitalWriteDirect(servo_right_pin, digitalRead(digitalReadDirect(button_start_pin)));      

/*    
  // motor acceleration
  act_micros = micros();
  delta = act_micros - last_motor_update;
  if (delta > 20000) {
    last_motor_update = act_micros;  
    for(i = 0; i < NUM_MOTORS; i++) {
      updateMotorPWM(i);  
      setMotorPWM(i, M_PWM_act[i], !button_stop);
    }
  }
*/  
  act_micros = micros();
  delta = act_micros - last_micros;
  if (delta > interval) {
    // save the last time 
    last_micros = act_micros;   
    dt = interval * 1e-6;

    prev_botton_stop = button_stop;    
    button_stop = !digitalRead(button_stop_pin);
    
    // Store last odo value
    for(i = 0; i < NUM_MOTORS; i++) {
      cli();
      odo[i] = encoder_delta[i];
      encoder_delta[i] = 0;
      sei();
      encoder_pos[i] += odo[i];
    }

    for(i = 0; i < NUM_MOTORS; i++) {
      w_odo[i] = RADS_PER_TICK * (odo[i] / dt);
    }

    for(i = 0; i < NUM_MOTORS; i++) {
      if (pid[i].active) {
        PWM_out[i] = calc_pid(pid[i], w_ref[i], w_odo[i]);
        setMotorReqPWM(i, PWM_out[i]);
        //Serial.print(PID_out[i]);
        //Serial.print(" ");
      }
      updateMotorPWM(i);  
      setMotorPWM(i, M_PWM_act[i], !button_stop);
    }

    // Start MetaPacket
    serial_channels.send('p', 9 + (((digitalRead(button_start_pin) << 1) | (!digitalRead(button_stop_pin))) << 8)); 

    // sendOdometry
    serial_channels.sendFloat('r', w_odo[0]); 
    serial_channels.sendFloat('s', w_odo[1]); 
    serial_channels.sendFloat('t', w_odo[2]); 
    serial_channels.sendFloat('u', w_odo[3]); 
    
    serial_channels.send('g', (digitalRead(button_start_pin) << 1) | digitalRead(button_stop_pin)); 
    serial_channels.send('k', delta); 

    serial_channels.send('v', encoder_pos[0]); 
    serial_channels.sendFloat('w', w_ref[0]); 

    // End MetaPacket
    serial_channels.send('p', 0); 

    decimate_led++;
    if (decimate_led > 20) {
      // if the heartbeat LED is off turn it on and vice-versa:
      if (led_state == LOW) {
        led_state = HIGH;
      } else {
        led_state = LOW;
      }
      digitalWrite(led_pin, led_state);
      //Serial.println(led_state);

      decimate_led = 0;
    }

    Serial.println();
  }
}
