#include <Arduino.h>

#include <TimerOne.h>
#include <Wire.h>
#include <Adafruit_PWMServoDriver.h>
#include <Adafruit_MotorShield.h>
#include "utility/Adafruit_MS_PWMServoDriver.h"

// called this way, it uses the default address 0x40
Adafruit_PWMServoDriver pwm = Adafruit_PWMServoDriver();
// you can also call it with a different address you want
//Adafruit_PWMServoDriver pwm = Adafruit_PWMServoDriver(0x41);
// you can also call it with a different address and I2C interface
//Adafruit_PWMServoDriver pwm = Adafruit_PWMServoDriver(&Wire, 0x40);

// Depending on your servo make, the pulse width min and max may vary, you 
// want these to be as small/large as possible without hitting the hard stop
// for max range. You'll have to tweak them as necessary to match the servos you
// have!
#define SERVOMIN  150 // this is the 'minimum' pulse length count (out of 4096)
#define SERVOMAX  600 // this is the 'maximum' pulse length count (out of 4096)

// Create the motor shield object with the default I2C address
Adafruit_MotorShield AFMS = Adafruit_MotorShield(); 
// Or, create it with a different I2C address (say for stacking)
// Adafruit_MotorShield AFMS = Adafruit_MotorShield(0x61); 

// Select which 'port' M1, M2, M3 or M4. In this case, M1
Adafruit_DCMotor* M[4];
// You can also make another motor on port M2
//Adafruit_DCMotor *myOtherMotor = AFMS.getMotor(2);

#include "channels.h"
#include "proj_types.h"

// 9 -> Servo Pulse Left * 
// 10 -> Servo Pulse Right *

// 2 <- ENC1_A
// 3 <- ENC1_B
// 4 <- ENC2_A
// 5 <- ENC2_B

// 6 <- Button Start * 
// 7 <- Button Stop *

// A11 <- Lidar motor voltage

// Metal Gearmotor 12VDC with 64 CPR Encoder 350rpm
// Color  - Function

// Black  - motor power
// Red    - motor power
// Blue   - Hall sensor Vcc (3.5 ? 20 V)
// Green  - Hall sensor GND
// Yellow - Hall sensor A output
// White  - Hall sensor B output

#define NUM_MOTORS 4  // ( 2-3 -> wheel motors)

#define WHEEL_RADIUS 0.01825
#define TICKS_RAD 19.10986
#define DT 0.040
#define WHEEL_DIST 0.0965

const int encoder1A_pin = 2;  // Motor1 encoder A input
const int encoder1B_pin = 3;  // Motor1 encoder B input
const int encoder2A_pin = 4;  // Motor2 encoder A input
const int encoder2B_pin = 5;  // Motor2 encoder B input

channels_t serial_channels;

byte encoder_state[NUM_MOTORS];
volatile uint32_t encoder_pos[NUM_MOTORS];
uint32_t old_encoder_pos[NUM_MOTORS];
uint16_t motor_timeout_count;

int M_PWM_value[NUM_MOTORS];

PID_t pid[NUM_MOTORS];
int odo[NUM_MOTORS];
int last_odo[NUM_MOTORS];

float w_odo[NUM_MOTORS];
float w_ref[NUM_MOTORS];
int PID_out[NUM_MOTORS];

int button_stop, prev_botton_stop;

void timer_interrupt(void)
{
  byte b, new_state;
  static int8_t encoder_table[16] = {0, 1, -1, 0,  -1, 0, 0, 1,  1, 0, 0, -1,  0, -1, 1, 0};

  // Put encoder 1 channels in the lowest bits
  new_state = (digitalRead(encoder1A_pin) ? (1 << 1) : 0) | (digitalRead(encoder1B_pin) ? 1 : 0);
  encoder_pos[1] += encoder_table[encoder_state[1] | new_state];
  encoder_state[1] = new_state << 2;

  // Put encoder 2 channels in the lowest bits
  new_state = (digitalRead(encoder2A_pin) ? (1 << 1) : 0) | (digitalRead(encoder2B_pin) ? 1 : 0);
  encoder_pos[2] += encoder_table[encoder_state[2] | new_state];
  encoder_state[2] = new_state << 2;
}


const int led_pin =  13;     // the number of the LED pin

const int button_start_pin = 6;
const int button_stop_pin = 7;

const int servo_left_pin = 9;
const int servo_right_pin = 10;


int servo_state = 0;

int servo_left_pulse = 1000;
int servo_right_pulse = 1500;
/*
void servo_interrupt(void)
{
  if (servo_state == 0) {
    Timer4.start(servo_left_pulse);
    digitalWrite(servo_left_pin, 1);  // Start Left Servo Pulse
    servo_state = 1;
  } else if (servo_state == 1) {
    digitalWrite(servo_left_pin, 0); // End Left Servo Pulse
    Timer4.start(servo_right_pulse);
    digitalWrite(servo_right_pin, 1);  // Start Right Servo Pulse
    servo_state = 2;
  } else if (servo_state == 2) {
    digitalWrite(servo_right_pin, 0); // End right Servo Pulse
    Timer4.start(20000 - (servo_left_pulse + servo_right_pulse));
    servo_state = 0;
  }
}
*/

// Variables will change:
int led_state = LOW;         // ledState used to set the LED
unsigned long last_millis = 0;        // will store last time LED was updated

// the follow variables is a long because the time, measured in miliseconds,
// will quickly become a bigger number than can be stored in an int.
long interval = 40;         // interval at which to blink (milliseconds)

bool led_on = false;

void swap(int& a, int& b)
{
 int t = a;
 a = b;
 b = t;
}


void setup_PIDs(void)
{
  pid[0].Kp = 1e-3;//19.2e-6;// / 12.0 * 5.0;
  pid[0].Ki = 5e-4;//1e-2;
  pid[0].Kd = 0; //1e-3;
  pid[0].Kf = 2.2e-3 / 12.0 * 5.0;
  pid[0].scale = 255;
  pid[0].dt = 0.04;
  pid[0].active = 1;

  pid[1].Kp = 2e-3;
  pid[1].Ki = 1e-2;
  pid[1].Kd = 0; //1e-3;
  pid[1].Kf = 2.1e-3;
  pid[1].scale = 255;
  pid[1].dt = 0.04;
  pid[1].active = 1;

  pid[2] = pid[1];
  pid[3] = pid[1];
}

float calc_pid(PID_t& pid, float w_ref, float w)
{
  float de, Se;
  pid.e = w_ref - w;
  pid.last_e = pid.e;
  pid.w = w;
  pid.w_ref = w_ref;

  Se = pid.Se + pid.e * pid.dt;
  de = (pid.e - pid.last_e) / pid.dt;

  // Remove integration for zero reference
  if (w_ref == 0) {
    Se = 0;
  }

  // Calc PID output
  pid.m = pid.Kp * pid.e + pid.Ki * Se + pid.Kd * de + pid.Kf * pid.w_ref;

  // Anti windup
  if (abs(pid.m) > 1) {
    pid.m = pid.m + pid.Ki * (pid.Se - Se);
  } else {
    pid.Se = Se;
  }

  if (pid.m > 1) {
    pid.m = 1;
  } else if (pid.m < -1) {
    pid.m = -1;
  }

  return pid.m * pid.scale;
}

void setMotorPWM(uint8_t num, int new_PWM, int mot_enable)
{
  if (num >= NUM_MOTORS) return;
  
  if (new_PWM > 255) new_PWM = 255;
  if (new_PWM < -255) new_PWM = -255;
  M_PWM_value[num] = new_PWM;
  //return;

  if (mot_enable) {
    if (new_PWM >= 0) {
      M[num]->setSpeed(new_PWM);
      M[num]->run(FORWARD); 
    } else {
      M[num]->setSpeed(abs(new_PWM));
      M[num]->run(BACKWARD);
    }
  } else {
    M[num]->run(RELEASE);
  }
}


// G
// H
// I - dt
// J
// K
// L
// M - Motors PWM
// N
// O
// P - PID Active
// Q - Set Encoder value
// R
// S - Servo Pulses
// T
// U
// V
// W
// X
// Y
// Z


// g - ping++
// h - LIDAR PID Kp
// i - LIDAR PID Ki
// j - LIDAR PID Kd
// k - LIDAR PID Kf
// l - MOTOR PID Kp
// m - MOTOR PID Ki
// n - MOTOR PID Kd
// o - MOTOR PID Kf
// p
// q
// r - W ref
// s
// t
// u
// v
// w
// x
// y
// z


void process_serial_packet(char channel, uint32_t value, channels_t& obj)
{
 byte i;

 if (channel == 'G') {

 } else if (channel == 'S')  {
   servo_left_pulse = (value >> 16) & 0xFFFF;
   servo_right_pulse = value & 0xFFFF;
   
 } else if (channel == 'M')  {
   int16_t pwm = value & 0xFFFF;
   uint8_t num_motor = (value >> 24) & 0x03;
   setMotorPWM(num_motor, pwm, !button_stop);
   motor_timeout_count = 0;
   for (i = 1; i < NUM_MOTORS; i++){
     pid[i].active = 0;
   }

 } else if (channel == 'P')  {
   for (i = 0; i < NUM_MOTORS; i++){
     pid[i].active = value;
   }

 } else if (channel == 'Q')  {
   for (i = 0; i < NUM_MOTORS; i++) {
       encoder_pos[i] = (value & 0xFFFF);
   }

 } else if (channel == 'r')  {
   int16_t ref = value & 0xFFFF;
   uint8_t num_motor = (value >> 24) & 0x03;
   w_ref[num_motor] = ref;
   for (i = 1; i < NUM_MOTORS; i++){
     pid[i].active = 1;
   }
 
 } else if (channel == 'h')  {
   pid[0].Kp = *((float*)(&value));

 } else if (channel == 'i')  {
   pid[0].Ki = *((float*)(&value));
   pid[0].Se = 0;
 
 } else if (channel == 'j')  {
   pid[0].Kd = *((float*)(&value));

 } else if (channel == 'k')  {
   pid[0].Kf = *((float*)(&value));

 } else if (channel == 'l')  {
   for (i = 1; i < NUM_MOTORS; i++)
     pid[i].Kp = *((float*)(&value));

 } else if (channel == 'm')  {
   for (i = 1; i < NUM_MOTORS; i++) {
     pid[i].Ki = *((float*)(&value));
     pid[i].Se = 0;
   }
 
 } else if (channel == 'n')  {
   for (i = 1; i < NUM_MOTORS; i++)
     pid[i].Kd = *((float*)(&value));

 } else if (channel == 'o')  {
   for (i = 1; i < NUM_MOTORS; i++)
     pid[i].Kf = *((float*)(&value));

 } else if (channel == 'I')  {
   interval = value;

 } else if (channel == 'g')  { // Ping
   obj.send(channel, value + 1);
   Serial.println(value + 1);
 }
}



void serial_write(uint8_t b)
{
  Serial.write(b);
}


void setup(void)
{
  int i;
  // set the digital pin as output:
  pinMode(led_pin, OUTPUT);      
  digitalWrite(led_pin, 0);      

  digitalWrite(servo_left_pin, 0);      
  pinMode(servo_left_pin, OUTPUT);      

  digitalWrite(servo_right_pin, 0);      
  pinMode(servo_right_pin, OUTPUT);      

  pinMode(button_start_pin, INPUT_PULLUP);
  pinMode(button_stop_pin, INPUT_PULLUP);
  button_stop = !digitalRead(button_stop_pin);
  prev_botton_stop = button_stop;

  pinMode(encoder1A_pin, INPUT_PULLUP);
  pinMode(encoder1B_pin, INPUT_PULLUP);
  pinMode(encoder2A_pin, INPUT_PULLUP);
  pinMode(encoder2B_pin, INPUT_PULLUP);
  

  //pinMode(PWM_pin, OUTPUT);
  //analogWrite(PWM_pin, PWM_out);      

  Timer1.attachInterrupt(timer_interrupt); // blinkLED to run every 0.15 seconds
  Timer1.initialize(50);

  //Timer3.attachInterrupt(timer_interrupt);
  //Timer3.start(50); // Calls every 50us

  //Timer4.attachInterrupt(servo_interrupt);
  //Timer4.start(20000); // Calls every 20 ms


  // initialize serial communication:
  Serial.begin(115200);

  serial_channels.init(process_serial_packet, serial_write);
  Serial.println("Serial Init");

  for (i = 0; i < 4; i++) {
    M[i] = AFMS.getMotor(i + 1);
  }

  AFMS.begin();  // create with the default frequency 1.6KHz
  //AFMS.begin(1000);  // OR with a different frequency, say 1KHz

  
  M[0]->setSpeed(150);
  M[0]->run(BACKWARD); 
  //M[0]->run(FORWARD); 

  setMotorPWM(1, 0, !button_stop); 
  setMotorPWM(2, 0, !button_stop); 
  setMotorPWM(3, 0, !button_stop); 
  
  setup_PIDs();

  pwm.begin();
  pwm.setPWMFreq(50);  // Analog servos run at ~50 Hz updates

  for (i = 0; i < 16; i++) {
    pwm.setPWM(i, 0, 40 + 20 * i);
  }

  last_millis = millis();
  Serial.println("Init Done");
}


void loop()
{
  byte i; 
  static int decimate_led  = 0;
  unsigned long act_millis, delta;

  if (Serial.available() > 0) {
    // read the oldest byte in the serial buffer:
    byte serialByte = Serial.read();
    serial_channels.StateMachine(serialByte);
    
    /*switch(serialByte){
      case '+':
        desired_lidar_speed += 1 << 6;
        Serial.println(desired_lidar_speed >> 6);
        break;
      case '-':
        desired_lidar_speed -= 1 << 6;
        Serial.println(desired_lidar_speed >> 6);
        break;
      case '0':
        //motor_enabled = 0;
        break;
      case '1':
        //motor_enabled = 1;
        break;
      
    }*/

    //Serial1.write(serialByte);
  }

  //digitalWriteDirect(servo_right_pin, digitalRead(digitalReadDirect(button_start_pin)));      
  
  act_millis = micros();

  delta = act_millis - last_millis;
  if (delta > interval * 1000) {
    // save the last time 
    last_millis = act_millis;   

    prev_botton_stop = button_stop;    
    button_stop = !digitalRead(button_stop_pin);
    
    // Store last odo value
    for(i = 1; i < NUM_MOTORS; i++) {
      last_odo[i] = odo[i];
      odo[i] = encoder_pos[i];
    }

    for(i = 1; i < NUM_MOTORS; i++) {
      //w_odo[i] = w_odo[i] * 0.9 + 0.1 *(odo[i] - last_odo[i]);
      w_odo[i] = (odo[i] - last_odo[i]) * (interval * 1000.0 / delta);
      PID_out[i] = calc_pid(pid[i], w_ref[i], w_odo[i]);
    }

    for(i = 1; i < NUM_MOTORS; i++) {
      if (pid[i].active) {
        setMotorPWM(i, PID_out[i], !button_stop);
        //Serial.print(PID_out[i]);
        //Serial.print(" ");
      }
    }

    // Start MetaPacket
    serial_channels.send('p', 9 + (((digitalRead(button_start_pin) << 1) | (!digitalRead(button_stop_pin))) << 8)); 

    // SendPWM
    uint32_t data32 = 0;

    for(i = 0; i < 3; i++) {
       if (M_PWM_value[i + 1] > 0) data32 |= 1 << i;
    }
    for(i = 0; i < 3; i++) {
       data32 = data32 << 8;
       data32 |= abs(M_PWM_value[i + 1]);
    }
    
    serial_channels.send('m', data32);

    // sendOdometry
    serial_channels.send('r', encoder_pos[1]); 
    serial_channels.send('s', encoder_pos[2]); 
    
    serial_channels.send('g', (digitalRead(button_start_pin) << 1) | digitalRead(button_stop_pin)); 
    //serial_channels.send('z', lidar_speed >> 6); 
    serial_channels.send('k', delta); 

    // End MetaPacket
    serial_channels.send('p', 0); 

    decimate_led++;
    if (decimate_led > 20) {
      // if the heartbeat LED is off turn it on and vice-versa:
      if (led_state == LOW) {
        led_state = HIGH;
      } else {
        led_state = LOW;
      }
      digitalWrite(led_pin, led_state);
      //Serial.println(led_state);

      decimate_led = 0;
    }

    Serial.println();
  }
}
